/*
	Simple threaded portscanner by Ralph, molnar_rcs@yahoo.com
	10 jan 2006
	Can be used and distributed freely without removing this header.
*/

/*
	Compile: gcc ptscan.c -o ptscan
	TODO: implement resume option based on the ptscan.dat file
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>

extern int errno;
#define MAX_THREAD 1000


//
int connectXT(int sock, struct sockaddr_in addr, int tout, int mode) 
{ 
  int res, valopt; 
  long arg; 
  fd_set myset; 
  struct timeval tv; 
  socklen_t lon; 

  
  if(mode == 1)
  {
	// Initialize the file descriptor set
	FD_ZERO (&myset);
	FD_SET (sock, &myset);
	
	// Initialize the timeout data structure
	tv.tv_sec = tout;
	tv.tv_usec = 0;
	
	// select returns 0 if timeout, 1 if input available, -1 if error.
	return (select (FD_SETSIZE, &myset, NULL, NULL, &tv));
  }
  	
  // Set non-blocking 
  arg = fcntl(sock, F_GETFL, NULL); 
  arg |= O_NONBLOCK; 
  fcntl(sock, F_SETFL, arg); 

  // Trying to connect with timeout 
  res = connect(sock, (struct sockaddr *)&addr, sizeof(addr)); 

  if (res < 0) 
  { 
	if (errno == EINPROGRESS) 
	{ 
		tv.tv_sec = tout; 
		tv.tv_usec = 0; 
		FD_ZERO(&myset); 
		FD_SET(sock, &myset); 
		if (select(sock+1, NULL, &myset, NULL, &tv) > 0) 
		{ 
			lon = sizeof(int); 
			getsockopt(sock, SOL_SOCKET, SO_ERROR, (void*)(&valopt), &lon); 
			if (valopt) 
			{
				// quick hack to avoid valopt being out of the error range
				// if valopt is not a known error we assume that it's a time out error
				// actually it's not a good idea to do this here
				if(valopt < EPERM || valopt > EMEDIUMTYPE) valopt = 110;
				return valopt; 
			}
		} 
		else 
		{
			// quick hack to avoid valopt being out of the error range
			// if valopt is not a known error we assume that it's a time out error
			if(valopt < EPERM || valopt > EMEDIUMTYPE) valopt = 110;
			return valopt;
		}
	} 
	else return errno; 
  } 
  arg = fcntl(sock, F_GETFL, NULL); 
  arg &= (~O_NONBLOCK); 
  fcntl(sock, F_SETFL, arg); 
  return 0;
} 


// write to socket func
int writesoc(int sock, int wfd)
{
 int i = 0;
 char ch;
 
 //rewind file
 lseek(wfd, 0,SEEK_SET); 
 
 // read file byte by byte
 while(read(wfd, &ch, 1))
 {

	 // write byte to sock
	 if(write(sock, &ch, 1) != 1)
	 {
		// error occured or no byte was written
		// assume that this is a socket error and return from the func
		return i;
	 }
	 i++;
 }
 return i;
 
}

// signal handler
void sh(int signum)
{
 fprintf(stderr, "SIGINT caught.\n");
 kill(0, SIGTERM);
}

// lock file
int wblock(int fd, short whence, int len, int start)
{
  struct flock lockStruct;

  memset(&lockStruct, 0, sizeof (struct flock));

  lockStruct.l_type = F_WRLCK;
  lockStruct.l_whence = whence;
  lockStruct.l_len = len;
  lockStruct.l_start = start;

  if(fcntl(fd, F_SETLK, &lockStruct) == -1)
   {
    //perror(name);
    return 0;
   }

  return 1;
}

//unlock file
int ublock(int fd, short whence, int len)
{
  struct flock lockStruct;

  memset(&lockStruct, 0, sizeof (struct flock));

  lockStruct.l_type = F_UNLCK;
  lockStruct.l_whence = whence;
  lockStruct.l_len = len;

  if(fcntl(fd, F_SETLK, &lockStruct) == -1)
   {
    //perror(name);
    return 0;
   }

  return 1;
}


//usage
void usage(char *argv)
{
 fprintf(stderr, "[ARGS] \n"
	" -b begin         : start IP in dotted format - (x.x.x.x)\n"
	" -e end           : end IP in dotted format - (x.x.x.x)\n"
	" -p port          : port to scan\n"
	" [ -t threads ]   : number of threads to run, max %d - default is 1\n"
	" [ -s sleep ]     : sleep time in seconds for each thread - default is 1\n"
	" [ -T C_timeout ] : timeout in seconds before aborting a connection attempt - default value is 5\n"
	" [ -l logFile ]   : output will be logged to logFile (this doesn't disable printing to the screen)\n"
	" [ -w file ]      : file that contains data that you want to write to the socket\n"
	"                    file can be binary or text\n"
	" [ -r nr ]        : reads at most `nr` bytes from the socket byte by byte, can be used without [-w]\n"
	" [ -R R_timeout ] : timeout in seconds before aborting a read attempt - default value is 5\n"
	"                    works only with -r or else is silently ignored\n"
	" [ -g string ]    : grep data read for socket for `string`\n"
	"                    can be used only with -r\n"
	"                    if <string> is found, `FOUND` will be displayed and `FOUND <string>` will be logged`\n"
	"                    if <string> is not found, NOTFND will be logged and/or displayed\n"	
	" [ -B ]           : run in background, printing to the screen will be disabled (errors, etc)\n"
	"                    this will only work if logfile specified\n"
	" [ -G ]           : specify this if you don't want `<string>` displayed in the log file\n"
	"                    useful when giving binary data as grep <string> and you don't want the logfile messed up\n\n"
	" Read the README file if any for more information on how reading, writing to socket is performed,\n"
	" why reading is aborted before `nr` are read, and so on.\n\n"
	
	, MAX_THREAD);
 exit(0);
}


//main
int main(int argc, char **argv)
{
 pid_t pid;
 
 int 	tc = 1, 
	port, 
	tmp_fd, 
	stime = 1,
	i = 1,
	args = 0,
	sock,
	csleep = 5,
	pipM,
	serror, 
	isbackgr = 0,
	wfd = -1,
	readcount = 0, rsleep = 5,
	grpfound = 0, hidegrep = 0,
	lfd = -1, logDS = 0;
	
 struct sockaddr_in addr;	
 
 char ch, *msg, *logfile = NULL, datach,
      *banner = "Simple threaded portscanner by Ralph, molnar_rcs@yahoo.com\n",
      *grpstring = NULL, 
      *wrdfile = NULL,
      *logdata;
      
 
 unsigned long   
 		start, 
		end; 
		
//FILE	*lfd = NULL;
 		
 //banner 
 fprintf(stderr, "%s", banner);
   
 //get options
 while ((ch = getopt(argc, argv, "t:b:e:s:p:T:l:w:r:R:g:BG")) != EOF)
 {
  	switch ((char) ch) 
	{
		case 't':
		{
 			// set proc count to 1st arg
 			tc = atoi(optarg);
			if(tc > MAX_THREAD)
			{
				fprintf(stderr, "Thred value #%d over the thread limit #%d.\n"
						"Edit MAX_THREAD and recompile if you want to use higher values.\n"
						"Aborting....\n", tc, MAX_THREAD);
				exit(0);
			}
			
			//create at least 1 scan thread
			if(tc < 1) tc = 1;
			//args++;			
 			break;
		}
		
		case 'b':
		{
 			start = ntohl(inet_addr(optarg));
			args += 1;			
 			break;
		}
		
		case 'e':
		{
			end = ntohl(inet_addr(optarg));
			args += 2;
			break;
		}
		
		case 's':
		{
			stime = atoi(optarg);
   			if(stime < 0)
    			{
				fprintf(stderr, "Sleep Time can't be negative\nAborting...\n");
				exit(0);
			}				
			break;
		}
		
		case 'T':
		{
			csleep = atoi(optarg);
   			if(csleep < 0)
    			{
				fprintf(stderr, "Connection Timeout can't be negative\nAborting...\n");
				exit(0);
			}
			
			if(csleep == 0)
			{
				fprintf(stderr, "Connection Timeout set to 0 is not adviced.\n");
				fprintf(stderr, "This usually generates improper scan results.\n");
			}				
			break;
		}
		
		case 'p':
		{
			port = atoi(optarg);
 			if(port > 65535 || port < 1)
  			{
				fprintf(stderr, "Port value most be between 1-65535.\n"
						"Aborting....\n");
				exit(0);
 			}			
			args += 4;
			break;
		}
		
		case 'l':
		{
			logfile = optarg;
			break;
		}
		
		case 'B':
		{
			isbackgr = 1;
			break;
		}
		
		case 'w':
		{
			wrdfile = optarg;			
			break;
		}
		
		case 'r':
		{
			readcount = atoi(optarg);
			break;
		}
		
		case 'R':
		{
			rsleep = atoi(optarg);
   			if(rsleep < 0)
    			{
				fprintf(stderr, "READ Timeout can't be negative\nAborting...\n");
				exit(0);
			}
			
			if(rsleep == 0)
			{
				fprintf(stderr, "READ Timeout set to 0 is not adviced.\n");
				fprintf(stderr, "This usually generates improper scan results.\n");
			}	
			break;
		}
		
		case 'g':
		{
			grpstring = optarg;
			break;
		}
		
		case 'G':
		{
			hidegrep = 1;
			break;
		}		
		
	}
 }
  
 if(args < 7)
 { 
   usage(argv[0]);
   exit(0);
 }

 // we don't want to let this happen
 // currently it will result in 1 single scan
 if(start > end)
 {
  fprintf(stderr, "IP error. Start IP is greater than end IP.\nAborting...");
  exit(0);
 }
 //open/create data file
 tmp_fd = open("ptscan.dat", O_RDWR | O_CREAT, 0644); 
 if (tmp_fd < 0)
  {
      perror(argv[0]);
      exit(-1);
  } 
 
 //write start + end to file to file
  if( (write(tmp_fd, &start, sizeof(start)) != sizeof(start)) || (write(tmp_fd, &end, sizeof(end)) != sizeof(end)) )
   {
    //shit happened to file, can't write our data. Either error or incomplete write, abort
    fprintf(stderr, "File error. Incomplete or bogus write.\n");
    fprintf(stderr, "ptscan.dat");
    close(tmp_fd);
    exit(-2);
   } 
 
  if(logfile != NULL)
  {
   	fprintf(stderr, "Logging enabled, creating logfile %s.\n", logfile);      
 	lfd = open(logfile, O_WRONLY | O_CREAT | O_TRUNC, 0644); 
	if (lfd == -1)
	{
		perror(logfile);
		fprintf(stderr, "Could not open log file, logging is DISABLED.\n");
		// is logging disabled and in background mode, exit
		if(isbackgr == 1)
		 {
		 	 fprintf(stderr, "Background mode: no logging, scan ABORTED.\n");
			 exit(0);
		 }      
	}
	else write(lfd, banner, strlen(banner));
  }  
 
 // check for write to socket first 
 if(wrdfile != NULL)
  {
   	fprintf(stderr, "Writing to socket enabled. Opening data file %s readonly.\n", wrdfile);      
 	wfd = open(wrdfile, O_RDONLY); 
	if (wfd < 0)
	{
		perror(wrdfile);
		fprintf(stderr, "Writing to socket disabled.\n");
	}
  }
 
 // check for read from sock and grep
 if(grpstring != NULL && readcount == 0) 
 {
 	fprintf(stderr, "Read (-r) not enabled or `nr` is 0.\n"
			"Grep for string DISABLED\n");
	// silently disable -G here, actualy the code make no diffrence
	// since grep is disabled, no attemp will be made to log the grepstring
	// but just to be sure, let's set it to 0 
	hidegrep = 0;
	grpstring = NULL;	
 }
   
 fprintf(stderr, "Creating thread(s)...\n");      
 
 // if in background mode, we fork here
 if(isbackgr == 1)
 {
 	pipM = fork();
	if(pipM == -1)
	 {
	  	perror("# Background mode:  Fork error. Aborting...");
	  	exit(0);
	 }
	if(pipM != 0)
	 {
	 	fprintf(stderr, "Forked into the background with PID [%d].\n", pipM);
		fprintf(stderr, "Use killall -9 %s to kill all threads.\n", argv[0]);
		exit(1); // parent exists silently
	 }	
}

 // loop: create threads
 i = 1;
 while (i <= tc)
  {
   	i++;
	//try to fork
	if((pid = fork()) == -1) 
	 {
	 	if(isbackgr != 1) perror("# Fork error");
		continue;
	 }
	
	// if child break loop... do some thinking, why?
	// Answer: childs would continue to create their own childs.... duh ;)
	if(pid == 0) break;
	
  }
 
 //only parent prints banner    
 if(pid != 0) 
   {
   	if(isbackgr != 1) 
		fprintf(stderr, "Threads ready.\n"
				"Starting scan. Sit back and relax, this may take a while.\n"
				"~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
   }
 else sleep(1);
 
 if(pid == 0)
 {
  do 
  { 
	// try to block file
	
	while(!wblock(tmp_fd, SEEK_SET, sizeof(start),0));

	//rewind the file
	lseek(tmp_fd, 0,SEEK_SET);
	
	//try to read
	if(read(tmp_fd, &start, sizeof(start)) != sizeof(start))
	{
		//shit happened, incomplete read
		if(isbackgr != 1) 
		 {
		  	fprintf(stderr, "File error. Incomplete or bogus read\n");
			perror("ptscan.dat");
		 	fprintf(stderr, "To avoid infinite loops, aborting process pid #%d\n", getpid());
		 }	
		exit(-3);
	}
	//incremet
	start++;
			
	//rewind
	lseek(tmp_fd, 0,SEEK_SET);
	
	//write back the updated data  
	if(write(tmp_fd, &start, sizeof(start)) != sizeof(start))
	{
		//shit happened, incomplete write
		if(isbackgr != 1)
		{ 
			fprintf(stderr, "File error. Incomplete or bogus write\n");
			perror("ptscan.dat");
			fprintf(stderr, "To avoid infinite loops, aborting process pid #%d\n", getpid());
		}
		
		exit(-4);
	}
	
	//unlock file	
	while(!ublock(tmp_fd, SEEK_SET, sizeof(int)));

	// do a check
	if(start > end + 1) break;		
	
	// skip 0 and 255
	if( (( (start - 1) & 0xff) == 255) || (( (start - 1) & 0xff) == 0) ) continue;
	
	// do da scan	
	if ((sock=socket(AF_INET,SOCK_STREAM,0)) == -1)
	{
		// might break here to avoid further errors
		// let's assume this is only a temp error only in this child
		// might use resolve_err here to print to stdout
		// resolve_err should cover all socket errors
		if(isbackgr != 1) perror("SockeT:");
		continue;
	}
		
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port); 	
	addr.sin_addr.s_addr =  htonl(start-1);

	// print/log
	
	if(isbackgr != 1) fprintf(stdout, "%s port %d: ", inet_ntoa(addr.sin_addr), port);
	if(lfd != -1) 
		{
			// if we grep for a string, we need 512 bytes + grepstring size for the logged message
			// if not we need at least + rcount bytes of memory... rcount is 0 if we din't want
			// to read from the socket
			if(grpstring != NULL) 
				logDS = 512 + strlen(grpstring);
			else 
				logDS = 512 + readcount;
			
			if( (logdata = malloc ( logDS ) ) != NULL)
				{ 
					bzero(logdata, logDS);
					sprintf(logdata, "%s port %d: ", inet_ntoa(addr.sin_addr) ,port);
				}
				
		}
	
	
	if ( (serror = connectXT(sock, addr, csleep, 0) ) == 0)
	{
		if(lfd != -1 && logdata != NULL) sprintf(logdata, "%sOPEN", logdata);
		if(isbackgr != 1) fprintf(stdout, "OPEN");
		
		// port is open check for read, write, grep
		
		//if write, write content of datafile to socket
		if(wfd != -1) 
		{
			i = writesoc(sock, wfd);			
			// log on the fly
			if(lfd != -1 && logdata != NULL) sprintf(logdata, "%s * WRITTEN (%d) bytes", logdata, i);
			if(isbackgr != 1) fprintf(stdout, " * WRITTEN (%d) bytes", i);
		}	
		
		// ivoke connectXT in read mode, we can store return value serror cuz 
		// the current value of serror has no use from here on
		if((serror = connectXT(sock, addr, rsleep, 1)) == 1)
		{
			//if read, read `nr` bytes from sock , if <= 0 skip step
			if(readcount > 0)
			{
				
	
				//read byte by byte
				i = 0; // used to count bytes read
				if(grpstring != NULL) 
				{
					// use msg to store read data (used also for error messages
					msg = malloc(readcount);
					bzero(msg, readcount);
					grpfound = 0;
				}
				else
				{
					if(lfd != -1 && logdata != NULL) sprintf(logdata, "%s * DATA ( ", logdata);
				}
	
				while (read(sock, &datach, 1) && i < readcount)
				{	
					i++;
					if(grpstring != NULL)
					{
						sprintf(msg,"%s%c", msg, datach);
						if (strstr(msg, grpstring))
						{
							// string found
							grpfound = 1;
							if(lfd != -1 && logdata != NULL) 
							{ 
								sprintf(logdata, "%s * FOUND", logdata);
								if(!hidegrep)  
									sprintf(logdata, "%s %s", logdata, grpstring);  
							}
							if(isbackgr != 1) fprintf(stdout, " * FOUND");
							break;
						}
					}
					else
					{					
						if(datach != '\r' )//|| datach != '\n')
						{
							// log on the fly	
							if(lfd != -1 && logdata != NULL) 
								sprintf(logdata, "%s%c", logdata, datach);
						}
						else break;
					}
				}
				//if(grpstring != NULL) free(msg);
				if(isbackgr != 1) 
				{
					if(grpfound == 0 && grpstring != NULL)
						fprintf(stdout, " * NOTFND * READ (%d) bytes", i);
					
					if(grpfound == 1)
						fprintf(stdout, " * READ (%d) bytes", i);
					
				}
				
				if(lfd != -1 && logdata != NULL) 
				{ 
					if(grpfound == 0 && grpstring != NULL) 
						sprintf(logdata, "%s * NOTFND * READ (%d) bytes",logdata, i);
					
					if(grpstring == NULL) 
						sprintf(logdata, "%s ) * READ (%d) bytes", logdata, i);
					
					if(grpfound == 1) 
						sprintf(logdata, "%s * READ (%d) bytes", logdata, i);
				}
			}
		}
		else 
		{
			if(serror == 0)
			{
				if(isbackgr != 1) fprintf(stdout, " * READ timeout");
				if(lfd != -1 && logdata != NULL)  sprintf(logdata, "%s * READ timeout", logdata);
			}
			else
			{
				if(isbackgr != 1) fprintf(stdout, " * READ %s", strerror(errno));
				if(lfd != -1 && logdata != NULL)  
					sprintf(logdata, "%s * READ %s", logdata, strerror(errno));
			}
		}
		
		// end of read, write, grep
		
		// log/print a newline
		if(isbackgr != 1) fprintf(stdout, "\n");
		if(lfd != -1 && logdata != NULL)  sprintf(logdata, "%s\n", logdata); 
			
		close(sock);
	}
	// connect timed out
	else 
	{
		if(lfd != -1 && logdata != NULL)  sprintf(logdata,  "%s%s\n", logdata, strerror(serror));//, strerror(serror));
		if(isbackgr != 1) fprintf(stdout, "%s\n", strerror(serror));
	}
	
	// write out log :)
	if(lfd != -1 && logdata != NULL)
		write(lfd, logdata, strlen(logdata));
	// time to sleep
	sleep(stime);

  } while (start <= end);
  //end scan 
  
  close(tmp_fd);
  
  // first child to exit closes logfile and/or data file used for write socket (if we have one)
  if(lfd != -1) close(lfd);
  if(wfd != -1) close(wfd);
  
 }
 else
 {
  // trap CTRL-C here
  signal(SIGINT, &sh);
  
  //wait for any child to exit
  wait(NULL);
  
  // small check, if we only have one scan thread, no sleep is needed
  if(tc > 1)
   {
  	if(isbackgr != 1) fprintf(stderr, "Sleeping %d seconds to let childs finish...\n", csleep);
  	sleep(csleep);
   }
  kill(0, SIGTERM);
 }
 //code never reached, just make compiler happy
 return 1; 
}

